#line 380 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include "desk.h"
#include "ptrbuf.h"

size_t      n_icons;
deskicon_t **icons;

#line 19 "desk.c"

static int decode_int32(unsigned char *str, unsigned char *end, uint32_t *V);
static int encode_int32(unsigned char *start, uint32_t V);
static int decode_int64(unsigned char *start,unsigned char *end,uint64_t *V);
static int encode_int64(unsigned char *start, uint64_t V);
static int encode_str(unsigned char *start, char *str);
static int decode_str(unsigned char *start, unsigned char *end, char **V);
static int encode_image
  (unsigned char *start, uint32_t w,uint32_t h,uint32_t *data);
static int decode_image(unsigned char *start, unsigned char *end,
                        uint32_t **Rimage);
static void encode_values(unsigned char **Rd,uint32_t *Rl,const char *fmt, ...);
static int decode_values
  (unsigned char *start,unsigned char *end, const char *fmt, ...);
static void encode_icon(deskicon_t *ico);
static void decode_icon(deskicon_t *ico);
static int write_force(int fd, void* data, int len);
static int read_force(int fd, void* data, int len);
static int get_number(int fd, uint32_t *V);
static int put_number(int fd, uint32_t V);
static void free_n_null(void *x);
#line 50 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

static int decode_int32(unsigned char *str, unsigned char *end, uint32_t *V)
{
  if (end-str<4) return -1;
  *V= str[0] | (str[1]<<8) | (str[2]<<16) | (str[3]<<24);
  return 4;
}

static int encode_int32(unsigned char *start, uint32_t V)
{
  int i;
  if (start)
    for(i=0;i<4;i++) start[i]= V>>(i*8);
  return 4;
}

static int decode_int64(unsigned char *start,unsigned char *end,uint64_t *V)
{
  int i;
  uint64_t K,Z;
  if (end-start<8) return -1;
  K= 0;
  for(i=0;i<8;i++) { Z= start[i]; Z<<= i*8; K |= Z; }
  *V= K;
  return 8;
}

static int encode_int64(unsigned char *start, uint64_t V)
{
  int i;
  if (start)
    for(i=0;i<8;i++) start[i]= V>>(i*8);
  return 8;
}

static int encode_str(unsigned char *start, char *str)
{
  int L,i;
  if (!str) str= "";
  L= strlen(str)+1;
  if (start) 
     for(i=0;i<L;i++) start[i]= str[i];
  return L;
}

static int decode_str(unsigned char *start, unsigned char *end, char **V)
{
  unsigned char *x;
  char *K;
  int L;
  for(x=start;x<end;x++) if (x[0]==0) break;
  if (x==end) return -1;
  L= x-start+1;
  K= malloc(L);
  memcpy(K,start,L);
  *V= K;
  return L;
}

static int encode_image
  (unsigned char *start, uint32_t w,uint32_t h,uint32_t *data)
{
  // I prefix the image data with the size of it so that the
  // decoder can easily allocate enough space without referrring
  // to the width and height members read.
  uint32_t sz;
  uint32_t D;
  int R;
  int i;
  sz= w*h;
  R= 0;
  if (!data) data= &D;
  D= 0xff0000;
  R+= encode_int32(start, sz);
  if (start)
     for(i=0;i<sz;i++)
     {
       R+= encode_int32(start+R, data[0]);
       if (data!=&D) data++;
     }
  return sz*4+4;
}

static int decode_image(unsigned char *start, unsigned char *end,
                        uint32_t **Rimage)
{
  uint32_t sz;
  uint32_t *img;
  int R,i;

  img= NULL;
  R= decode_int32(start,end, &sz);
  if (R<0) goto fail;
  *Rimage= img= malloc(sz*4);
  start+= R;
  for(i=0;i<sz;i++)
  {
    R= decode_int32(start,end,img+i);
    if (R<0) goto fail;
    start+= R;
  }
  return sz*4+4;
fail:
  if (img) free(img);
  *Rimage= NULL;
  return -1;
}

static void encode_values(unsigned char **Rd,uint32_t *Rl,const char *fmt, ...)
{
  va_list ap;
  int i,R;
  size_t L;
  unsigned char *d;
  uint32_t I1,I2,*IP1;
  d= NULL;

again:
  L= 0;
  va_start(ap,fmt);
  for(i=0;fmt[i];i++)
  {
   R= -1;
   switch(fmt[i]) 
   {
    case 'w': R= encode_int32(d, va_arg(ap,uint32_t)); break;
    case 'd': R= encode_int64(d, va_arg(ap,uint64_t)); break;
    case 's': R= encode_str(d, va_arg(ap,char*)); break;
    case 'i':
       I1= va_arg(ap,uint32_t);
       I2= va_arg(ap,uint32_t);
       IP1= va_arg(ap,uint32_t*);
       R= encode_image(d, I1, I2,IP1);
       break;
   }
   if (R<0) break;
   if (d) d+= R;
   L+= R;
  }
  va_end(ap);
  if (!d)
  {
   *Rd= d= malloc(L);
   *Rl= L;
   goto again;
  }
}

static int decode_values
  (unsigned char *start,unsigned char *end, const char *fmt, ...)
{
  int i,R;
  va_list ap;
  va_start(ap,fmt);
  for(i=0;fmt[i];i++)
  {
   R= -1;
   switch(fmt[i])
   {
    case 'w': R= decode_int32(start,end, va_arg(ap,uint32_t*)); break;
    case 'd': R= decode_int64(start,end, va_arg(ap,uint64_t*)); break;
    case 's': R= decode_str(start,end, va_arg(ap,char **)); break;
    case 'i': R= decode_image(start, end, va_arg(ap,uint32_t**)); break;
   }
   if (R<0) { va_end(ap); return -1; }
   start+= R;
  }
  va_end(ap);
  return 0;
}

static void encode_icon(deskicon_t *ico)
{
  if (ico->encoded) return ;
  encode_values(&ico->encoded, &ico->encoded_l, 
   "wwwwwissw", 
    ico->type,
     ico->x,
      ico->y,
       ico->w,
        ico->h,
         ico->w,ico->h, ico->ico,
          ico->cmd, 
           ico->fn, 
            ico->mtime);
}

static void decode_icon(deskicon_t *ico)
{
  if (!ico->encoded) return ;
  decode_values(ico->encoded, ico->encoded+ico->encoded_l,
   "wwwwwissw", 
    &ico->type,
     &ico->x,
      &ico->y,
       &ico->w,
        &ico->h,
         &ico->ico,
          &ico->cmd, 
           &ico->fn, 
            &ico->mtime);

 // printf("icon: (%d,%d)[%dx%d]: %s\n", ico->x, ico->y, ico->w, ico->h, ico->fn);
}

#line 280 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

static int write_force(int fd, void* data, int len)
{
  int R;
  while(len)
  {
    R= write(fd, data, len);
    if (R<0)
    { if (errno==EINTR) continue;
      fprintf(stderr,"error writing file: %s\n", strerror(errno));
      return 1; }
    data+= R;
    len-= R;
  }
  return 0;
}


static int read_force(int fd, void* data, int len)
{
  int R;
  while(len)
  {
    R= read(fd, data, len);
    if (R<0)
    { if (errno==EINTR) continue;
      fprintf(stderr,"error reading file: %s\n", strerror(errno));
      return 1; }
    if (R==0) { fprintf(stderr,"unexpected EOF in file\n"); return 1; }
    data+= R;
    len-= R;
  }
  return 0;
}

#line 349 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

static int get_number(int fd, uint32_t *V)
{
  unsigned char nr[4];
  if (read_force(fd, nr,4)) return 1;
  decode_int32(nr, nr+4, V);
  return 0;
}

static int put_number(int fd, uint32_t V)
{
  unsigned char nr[4];
  encode_int32(nr, V);
  if (write_force(fd, nr, 4)) return 1;
  return 0;
}

#line 483 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

static void free_n_null(void *x)
{
  if (x) free(x);
}

#line 256 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

void write_desk_file(char *fn)
{
  int fd;
  int i;

  fd= open(fn, O_WRONLY | O_CREAT | O_TRUNC, 0600);
  if (fd<0)
  { fprintf(stderr,"can't open %s: %s\n", fn, strerror(errno));
    return ; } 

  if (put_number(fd, n_icons)) goto done;

  for(i=0;i<n_icons;i++)
  {
    encode_icon(icons[i]);
    if (put_number(fd, icons[i]->encoded_l)) goto done;
    if (write_force(fd, icons[i]->encoded, icons[i]->encoded_l)) goto done;
  }
done:
  close(fd);
}

#line 316 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

int read_desk_file(char *fn)
{
  int fd;
  uint32_t nico;
  deskicon_t *ico;
  int i;

  fd= open(fn,O_RDONLY);
  if (fd<0)
  { fprintf(stderr,"can't open %s: %s\n",fn, strerror(errno)); 
    return 1; } 
  if (get_number(fd, &nico) || nico==0 || nico>=500) goto fail;

  icons= calloc(nico,sizeof(icons[0]));
  for(i=0;i<nico;i++)
  {
    ico= icons[i]= calloc(1,sizeof(ico[0]));
    if (get_number(fd, &ico->encoded_l)) goto done;
    ico->encoded= malloc(ico->encoded_l);
    if (read_force(fd, ico->encoded, ico->encoded_l)) goto done;
    decode_icon(ico);
  }
done:
  n_icons= i;
  close(fd);
  return 0;
fail:
  close(fd);
  return 1;
}

#line 418 "/home/dodo/tmp/mudesk-1.0/desk/desk.u"

int scan_desktop_dir(char *dn, int *Rnfil, deskfile_t ***Rfil)
{
  DIR *dir;
  deskfile_t *fil;
  struct dirent *de;
  ptrbuf_t *files;
  char buf[2000];
  size_t L;

  dir= opendir(dn);
  if (!dir) return 1;
  files= ptrbuf_new(50);
  while((de=readdir(dir)))
  {
    if (de->d_name[0]=='.') continue;
    snprintf(buf,sizeof(buf),"%s/%s",dn,de->d_name);
    buf[sizeof(buf)-1]= 0;
    if ((fil= stat_file(buf)))
      ptrbuf_put(files, fil);
  }
  closedir(dir);

  ptrbuf_free(files, (void***) Rfil, &L);
  *Rnfil= L;
  return 0;
}

deskfile_t *stat_file(char *fn)
{
  deskfile_t *fil;
  int type;
  struct stat sta;
  if (stat(fn,&sta)) return NULL;
  if (S_ISDIR(sta.st_mode))
    type= 'D';
  else if (S_ISREG(sta.st_mode))
    type= 'F';
  else return NULL;
  fil= calloc(1,sizeof(*fil));
  fil->name= strdup(fn);
  fil->type= type;
  fil->mtime= sta.st_mtime;
  return fil;
}



void deskicon_free(deskicon_t *ic)
{
  free_n_null(ic->ico);
  free_n_null(ic->cmd);
  free_n_null(ic->fn);
  free_n_null(ic->encoded);
  free(ic);
}

void deskfile_free(deskfile_t *fi)
{
  free_n_null(fi->name);
  free(fi);
}


